@@ -14,7 +14,9 @@ import (
 	"github.com/canonical/go-dqlite"
 	"github.com/canonical/go-dqlite/client"
 	"github.com/canonical/go-dqlite/driver"
+	"github.com/canonical/go-dqlite/internal/protocol"
 	"github.com/pkg/errors"
+	"golang.org/x/sync/semaphore"
 )
 
 // App is a high-level helper for initializing a typical dqlite-based Go
@@ -30,10 +32,12 @@ type App struct {
 	nodeBindAddress string
 	listener        net.Listener
 	tls             *tlsSetup
+	dialFunc        client.DialFunc
 	store           client.NodeStore
 	driver          *driver.Driver
 	driverName      string
 	log             client.LogFunc
+	ctx             context.Context
 	stop            context.CancelFunc // Signal App.run() to stop.
 	proxyCh         chan struct{}      // Waits for App.proxy() to return.
 	runCh           chan struct{}      // Waits for App.run() to return.
@@ -50,6 +54,17 @@ func New(dir string, options ...Option) (app *App, err error) {
 		option(o)
 	}
 
+	var nodeBindAddress string
+	if o.Conn != nil {
+		listener, err := net.Listen("unix", o.UnixSocket)
+		if err != nil {
+			return nil, fmt.Errorf("Failed to autobind unix socket: %w", err)
+		}
+
+		nodeBindAddress = listener.Addr().String()
+		listener.Close()
+	}
+
 	// List of cleanup functions to run in case of errors.
 	cleanups := []func(){}
 	defer func() {
@@ -70,7 +85,9 @@ func New(dir string, options ...Option) (app *App, err error) {
 	}
 	if !infoFileExists {
 		if o.Address == "" {
-			o.Address = defaultAddress()
+			if o.Address, err = defaultAddress(); err != nil {
+				return nil, err
+			}
 		}
 		if len(o.Cluster) == 0 {
 			info.ID = dqlite.BootstrapID
@@ -143,9 +160,11 @@ func New(dir string, options ...Option) (app *App, err error) {
 	}
 
 	// Start the local dqlite engine.
-	var nodeBindAddress string
+	ctx, stop := context.WithCancel(context.Background())
 	var nodeDial client.DialFunc
-	if o.TLS != nil {
+	if o.Conn != nil {
+		nodeDial = extDialFuncWithProxy(ctx, o.Conn.dialFunc)
+	} else if o.TLS != nil {
 		nodeBindAddress = fmt.Sprintf("@dqlite-%d", info.ID)
 
 		// Within a snap we need to choose a different name for the abstract unix domain
@@ -156,7 +175,7 @@ func New(dir string, options ...Option) (app *App, err error) {
 			nodeBindAddress = fmt.Sprintf("@snap.%s.dqlite-%d", snapInstanceName, info.ID)
 		}
 
-		nodeDial = makeNodeDialFunc(o.TLS.Dial)
+		nodeDial = makeNodeDialFunc(ctx, o.TLS.Dial)
 	} else {
 		nodeBindAddress = info.Address
 		nodeDial = client.DefaultDialFunc
@@ -170,9 +189,11 @@ func New(dir string, options ...Option) (app *App, err error) {
 		dqlite.WithSnapshotParams(o.SnapshotParams),
 	)
 	if err != nil {
+		stop()
 		return nil, fmt.Errorf("create node: %w", err)
 	}
 	if err := node.Start(); err != nil {
+		stop()
 		return nil, fmt.Errorf("start node: %w", err)
 	}
 	cleanups = append(cleanups, func() { node.Close() })
@@ -181,10 +202,13 @@ func New(dir string, options ...Option) (app *App, err error) {
 	driverDial := client.DefaultDialFunc
 	if o.TLS != nil {
 		driverDial = client.DialFuncWithTLS(driverDial, o.TLS.Dial)
+	} else if o.Conn != nil {
+		driverDial = o.Conn.dialFunc
 	}
 
 	driver, err := driver.New(store, driver.WithDialFunc(driverDial), driver.WithLogFunc(o.Log))
 	if err != nil {
+		stop()
 		return nil, fmt.Errorf("create driver: %w", err)
 	}
 	driverIndex++
@@ -192,14 +216,19 @@ func New(dir string, options ...Option) (app *App, err error) {
 	sql.Register(driverName, driver)
 
 	if o.Voters < 3 || o.Voters%2 == 0 {
+		stop()
 		return nil, fmt.Errorf("invalid voters %d: must be an odd number greater than 1", o.Voters)
 	}
 
 	if o.StandBys%2 == 0 {
+		stop()
 		return nil, fmt.Errorf("invalid stand-bys %d: must be an odd number", o.StandBys)
 	}
 
-	ctx, stop := context.WithCancel(context.Background())
+	if runtime.GOOS != "linux" && nodeBindAddress[0] == '@' {
+		// Do not use abstract socket on other platforms and left trim "@"
+		nodeBindAddress = nodeBindAddress[1:]
+	}
 
 	if runtime.GOOS != "linux" && nodeBindAddress[0] == '@' {
 		// Do not use abstract socket on other platforms and left trim "@"
@@ -213,10 +242,12 @@ func New(dir string, options ...Option) (app *App, err error) {
 		node:            node,
 		nodeBindAddress: nodeBindAddress,
 		store:           store,
+		dialFunc:        driverDial,
 		driver:          driver,
 		driverName:      driverName,
 		log:             o.Log,
 		tls:             o.TLS,
+		ctx:             ctx,
 		stop:            stop,
 		runCh:           make(chan struct{}, 0),
 		readyCh:         make(chan struct{}, 0),
@@ -240,6 +271,28 @@ func New(dir string, options ...Option) (app *App, err error) {
 
 		cleanups = append(cleanups, func() { listener.Close(); <-proxyCh })
 
+	} else if o.Conn != nil {
+		go func() {
+			for {
+				remote := <-o.Conn.acceptCh
+
+				// Write the status line and upgrade header by hand since w.WriteHeader() would fail after Hijack().
+				data := []byte("HTTP/1.1 101 Switching Protocols\r\nUpgrade: dqlite\r\n\r\n")
+				n, err := remote.Write(data)
+				if err != nil || n != len(data) {
+					remote.Close()
+					panic(fmt.Errorf("Failed to write connection header: %w", err))
+				}
+
+				local, err := net.Dial("unix", nodeBindAddress)
+				if err != nil {
+					remote.Close()
+					panic(fmt.Errorf("Failed to connect to bind address %q: %w", nodeBindAddress, err))
+				}
+
+				go proxy(app.ctx, remote, local, nil)
+			}
+		}()
 	}
 
 	go app.run(ctx, o.RolesAdjustmentFrequency, joinFileExists)
@@ -418,7 +471,7 @@ func (a *App) Client(ctx context.Context) (*client.Client, error) {
 // Proxy incoming TLS connections.
 func (a *App) proxy() {
 	wg := sync.WaitGroup{}
-	ctx, cancel := context.WithCancel(context.Background())
+	ctx, cancel := context.WithCancel(a.ctx)
 	for {
 		client, err := a.listener.Accept()
 		if err != nil {
@@ -598,32 +651,52 @@ again:
 // RolesChanges object.
 func (a *App) makeRolesChanges(nodes []client.NodeInfo) RolesChanges {
 	state := map[client.NodeInfo]*client.NodeMetadata{}
-
 	for _, node := range nodes {
 		state[node] = nil
-		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
-		defer cancel()
-
-		cli, err := client.New(ctx, node.Address, a.clientOptions()...)
-		if err == nil {
-			metadata, err := cli.Describe(ctx)
-			if err == nil {
-				state[node] = metadata
-			}
-			cli.Close()
-		}
 	}
 
+	var (
+		mtx     sync.Mutex     // Protects state map
+		wg      sync.WaitGroup // Wait for all probes to finish
+		nProbes = runtime.NumCPU()
+		sem     = semaphore.NewWeighted(int64(nProbes)) // Limit number of parallel probes
+	)
+
+	for _, node := range nodes {
+		wg.Add(1)
+		// sem.Acquire will not block forever because the goroutines
+		// that release the semaphore will eventually timeout.
+		if err := sem.Acquire(context.Background(), 1); err != nil {
+			a.warn("Failed to acquire semaphore: %v", err)
+			wg.Done()
+			continue
+		}
+		go func(node protocol.NodeInfo) {
+			defer wg.Done()
+			defer sem.Release(1)
+			ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
+			defer cancel()
+
+			cli, err := client.New(ctx, node.Address, a.clientOptions()...)
+			if err == nil {
+				metadata, err := cli.Describe(ctx)
+				if err == nil {
+					mtx.Lock()
+					state[node] = metadata
+					mtx.Unlock()
+				}
+				cli.Close()
+			}
+		}(node)
+	}
+
+	wg.Wait()
 	return RolesChanges{Config: a.roles, State: state}
 }
 
 // Return the options to use for client.FindLeader() or client.New()
 func (a *App) clientOptions() []client.Option {
-	dial := client.DefaultDialFunc
-	if a.tls != nil {
-		dial = client.DialFuncWithTLS(dial, a.tls.Dial)
-	}
-	return []client.Option{client.WithDialFunc(dial), client.WithLogFunc(a.log)}
+	return []client.Option{client.WithDialFunc(a.dialFunc), client.WithLogFunc(a.log)}
 }
 
 func (a *App) debug(format string, args ...interface{}) {
